from mathics.core.symbols import Symbol

# DirectedEdge and UndirectedEdge are process in the parser of mathics-core
# as symbols in System`.
SymbolDirectedEdge = Symbol("System`DirectedEdge")
SymbolGraph = Symbol("Pymathics`Graph")
SymbolGraphBox = Symbol("Pymathics`GraphBox")
SymbolUndirectedEdge = Symbol("System`UndirectedEdge")
SymbolTwoWayRule = Symbol("System`TwoWayRule")
